# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Connectome extraction module."""

import os

import json
import numpy as np

from connectome_manipulator import log, profiler
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation


class ConnectomeExtraction(Manipulation):
    """Connectome manipulation class for extracting connections:

    Extracts a subset of connections within a cell target, i.e., keeps
    only connections within that target and removed all other connections
    from the connectome. The manipulation can be applied through the :func:`apply`
    method. The nodes (neuron) table is always kept unchanged!
    """

    @profiler.profileit(name="conn_extraction")
    def apply(self, split_ids, target_name=None, node_sets_file=None, **kwargs):
        """Applies an extraction of connections within a given cell target.

        Args:
            split_ids (list-like): List of neuron IDs that are part of the current data split; will be automatically provided by the manipulator framework
            target_name (str): Cell target name (as defined in SONATA node sets file) to extract connectome from, i.e, keeping only connections within that target; an empty connectome will be returned if no `target_name` is provided
            node_sets_file (str): Optional file path to alternative SONATA node sets file (.json) containing `node_id` entries; can be provided in case the cell target is not part of the circuit's intrinsic node sets.
            **kwargs: Additional keyword arguments - Not used

        Note:
            Input/output edges (synapse) tables are accessed through the ``writer`` object:

            * Loading input edges: ``edges_table = self.writer.to_pandas()``
            * Writing output edges: ``self.writer.from_pandas(edges_table_manip)``
        """
        # pylint: disable=arguments-differ
        edges_table = self.writer.to_pandas()
        if target_name is None:
            log.info("No target name provided, returning empty connectome!")
            self.writer.from_pandas(edges_table.loc[[]].copy())
            return

        # Load cell targets
        assert (
            self.nodes[0] is self.nodes[1]
        ), "ERROR: Only one source/target node population supported!"
        # pylint: disable=W0212
        if target_name in self.nodes[0]._node_sets.content:
            target_gids = self.nodes[0].ids(target_name)
        else:
            log.log_assert(
                node_sets_file is not None,
                f'Target "{target_name}" unknown, node sets file required!',
            )

            # Load node sets file
            log.log_assert(
                os.path.exists(node_sets_file), f'Node sets file "{node_sets_file}" not found!'
            )
            with open(node_sets_file, "r") as f:
                node_sets = json.load(f)

            log.log_assert(
                target_name in node_sets.keys(),
                f'Target "{target_name}" not found in "{node_sets_file}"!',
            )
            log.log_assert(
                "node_id" in node_sets[target_name],
                f'Node IDs for target "{target_name}" not found in "{node_sets_file}"!',
            )
            target_gids = node_sets[target_name]["node_id"]
        log.info(f'Found {len(target_gids)} GIDs for target "{target_name}" to extract')

        # Extract connectome
        syn_sel_idx = np.logical_and(
            np.isin(edges_table["@source_node"], target_gids),
            np.isin(edges_table["@target_node"], target_gids),
        )
        edges_table_manip = edges_table[syn_sel_idx].copy()

        # TESTING/DEBUGGING #
        log.log_assert(
            not np.any(edges_table_manip.sum(1) == 0),
            f"Empty edges table entries found!\ntarget_gids={target_gids}\np.sum(syn_sel_idx)={np.sum(syn_sel_idx)}\nedges_table_manip.size={edges_table_manip.size}",
        )
        # ################# #

        self.writer.from_pandas(edges_table_manip)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Test Dask auxiliary functions"""

from connectome_manipulator.processing import BatchInfo


def test_grouping():
    """Test grouping of batches for target payload"""
    batches = [
        BatchInfo(3, {}, [0]),
        BatchInfo(29, {}, [1]),
        BatchInfo(10, {}, [2]),
        BatchInfo(10, {}, [3]),
        BatchInfo(10, {}, [4]),
        BatchInfo(3, {}, [5]),
    ]

    groups = BatchInfo.group_batches(batches, 30)
    assert len(groups) == 4

    assert len(groups[0]) == 1
    assert len(groups[1]) == 1
    assert len(groups[2]) == 3
    assert len(groups[3]) == 1

    assert [b.node_ids[0] for b in groups[0]] == [0]
    assert [b.node_ids[0] for b in groups[1]] == [1]
    assert [b.node_ids[0] for b in groups[2]] == [2, 3, 4]
    assert [b.node_ids[0] for b in groups[3]] == [5]



# This file is part of connectome-manipulator.

"""Collection of function for flexible nodes/edges access, to be used by model building and manipulation operations"""

import numpy as np
import pandas as pd

import libsonata
from bluepysnap.sonata_constants import Node, DYNAMICS_PREFIX
from bluepysnap.utils import add_dynamic_prefix
from bluepysnap.utils import euler2mat, quaternion2mat
from sklearn.model_selection import KFold

from connectome_manipulator import log


def property_names(nodes):
    """Get all property names for a population"""
    population = nodes.to_libsonata
    return set(population.attribute_names) | set(
        add_dynamic_prefix(population.dynamics_attribute_names)
    )


def _ids_to_ranges(ids):
    """Convert a list of IDs to libsonata.Selection style ranges"""
    return libsonata.Selection(ids).ranges


def get_nodes(nodes, selection=None):
    """Get a pandas table of all nodes and their properties, optionally narrowed by a selection"""
    population = nodes.to_libsonata
    categoricals = population.enumeration_names

    if selection is None:
        selection = nodes.select_all()
    result = pd.DataFrame(index=selection.flatten())

    for attr in sorted(population.attribute_names):
        if attr in categoricals:
            enumeration = np.asarray(population.get_enumeration(attr, selection))
            values = np.asarray(population.enumeration_values(attr))
            # if the size of `values` is large enough compared to `enumeration`, not using
            # categorical reduces the memory usage.
            if values.shape[0] < 0.5 * enumeration.shape[0]:
                result[attr] = pd.Categorical.from_codes(enumeration, categories=values)
            else:
                result[attr] = values[enumeration]
        else:
            result[attr] = population.get_attribute(attr, selection)
    for attr in sorted(add_dynamic_prefix(population.dynamics_attribute_names)):
        result[attr] = population.get_dynamics_attribute(attr.split(DYNAMICS_PREFIX)[1], selection)
    return result


def orientations(nodes, node_sel=None):
    """Node orientation(s) as a list of numpy arrays.

    Args:
        nodes: the node set for which we want to get the rotation matrices
        node_sel: (optional) a libsonata Selection to narrow our selection

    Returns:
        numpy.ndarray:
            A list of 3x3 rotation matrices for the given node set and selection.
    """
    # need to keep this quaternion ordering for quaternion2mat (expects w, x, y , z)
    props = np.array(
        [Node.ORIENTATION_W, Node.ORIENTATION_X, Node.ORIENTATION_Y, Node.ORIENTATION_Z]
    )
    props_mask = np.isin(props, list(property_names(nodes)))
    orientation_count = np.count_nonzero(props_mask)
    if orientation_count == 4:
        trans = quaternion2mat
    elif orientation_count in [1, 2, 3]:
        raise ValueError(
            "Missing orientation fields. Should be 4 quaternions or euler angles or nothing"
        )
    else:
        # need to keep this rotation_angle ordering for euler2mat (expects z, y, x)
        props = np.array(
            [
                Node.ROTATION_ANGLE_Z,
                Node.ROTATION_ANGLE_Y,
                Node.ROTATION_ANGLE_X,
            ]
        )
        props_mask = np.isin(props, list(property_names(nodes)))
        trans = euler2mat
    result = get_nodes(nodes, node_sel)
    if props[props_mask].size:
        result = result[props[props_mask]]

    def _get_values(prop):
        """Retrieve prop from the result Dataframe/Series."""
        if isinstance(result, pd.Series):
            return [result.get(prop, 0)]
        return result.get(prop, np.zeros((result.shape[0],)))

    args = [_get_values(prop) for prop in props]
    return trans(*args)


def get_enumeration_list(pop, column):
    """Takes a node population and column name and returns a list to values."""
    raw_pop = pop.to_libsonata
    if column in raw_pop.enumeration_names:
        return raw_pop.enumeration_values(column)
    return sorted(np.unique(raw_pop.get_attribute(column, raw_pop.select_all())))


def get_enumeration_map(pop, column):
    """Takes a node population and column name and returns a dictionary that maps values to indices."""
    raw_pop = pop.to_libsonata
    if column in raw_pop.enumeration_names:
        return {key: idx for idx, key in enumerate(raw_pop.enumeration_values(column))}
    return {
        key: idx
        for idx, key in enumerate(
            sorted(np.unique(raw_pop.get_attribute(column, raw_pop.select_all())))
        )
    }


def get_attribute(pop, column, ids):
    """Get the attribute values for `column` from population `pop` for node IDs `ids`."""
    raw_pop = pop.to_libsonata
    return raw_pop.get_attribute(column, libsonata.Selection(ids))


def get_enumeration(pop, column, ids=None):
    """Get the raw enumeration values for `column` from population `pop` for node IDs `ids`."""
    raw_pop = pop.to_libsonata
    if ids is None:
        ids = raw_pop.select_all()
    else:
        ids = libsonata.Selection(ids)
    if column in raw_pop.enumeration_names:
        return raw_pop.get_enumeration(column, ids)
    mapping = get_enumeration_map(pop, column)
    return np.array([mapping[v] for v in raw_pop.get_attribute(column, ids)])


def get_node_ids(nodes, sel_spec, split_ids=None):
    """Returns list of selected node IDs of given nodes population.

    nodes ... NodePopulation
    sel_spec ... Node selection specifier, as accepted by nodes.ids(group=sel_spec).
                 In addition, if sel_spec is a dict, 'node_set': '<node_set_name>'
                 can be specified in combination with other selection properties.
    split_ids ... Node IDs to filter the selection by, either as an array or a
                  libsonata.Selection
    """
    pop = nodes.to_libsonata
    enumeration_names = pop.enumeration_names
    if split_ids is None:
        sel_ids = pop.select_all()
    else:
        sel_ids = libsonata.Selection(split_ids)
    if isinstance(sel_spec, dict):
        sel_group = sel_spec.copy()
        node_set = sel_group.pop("node_set", None)

        selection = None
        for sel_k, sel_v in sel_group.items():
            if sel_k in enumeration_names:
                if isinstance(sel_v, list):  # Merge multiple selections
                    sel_idx = [pop.enumeration_values(sel_k).index(_v) for _v in sel_v]
                    sel_prop = np.isin(pop.get_enumeration(sel_k, sel_ids), sel_idx)
                else:  # Single selection
                    sel_idx = pop.enumeration_values(sel_k).index(sel_v)
                    sel_prop = pop.get_enumeration(sel_k, sel_ids) == sel_idx
            else:
                if isinstance(sel_v, list):  # Merge multiple selections
                    sel_prop = np.isin(pop.get_attribute(sel_k, sel_ids), sel_v)
                else:  # Single selection
                    sel_prop = pop.get_attribute(sel_k, sel_ids) == sel_v
            if selection is None:
                selection = sel_prop
            else:
                selection &= sel_prop
        # selection is not of all nodes (starting with node id 0), but a generic subset specified by sel_ids
        if len(sel_ids.ranges) == 1:
            # First turn selection array into an index array then
            # because we filtered contiguous ids with a simple offset, just shift by the first node id
            gids = np.atleast_1d(selection).nonzero()[0] + sel_ids.ranges[0][0]
        else:
            # for more complex cases, fully resolve the node-preselection
            gids = sel_ids.flatten().astype(np.int64)[selection]

        if node_set is not None:
            log.log_assert(isinstance(node_set, str), "Node set must be a string!")
            if selection is None:  # Nothing else selected
                gids = nodes.ids(node_set)
                if split_ids is not None:
                    gids = np.intersect1d(gids, sel_ids.flatten().astype(np.int64))
            else:  # Otherwise, intersect with selection
                gids = np.intersect1d(gids, nodes.ids(node_set))
    else:
        gids = nodes.ids(sel_spec)
        if split_ids is not None:
            gids = np.intersect1d(gids, sel_ids.flatten().astype(np.int64))

    return gids


def get_nodes_population(circuit, popul_name=None):
    """Select default nodes population. Optionally, the population name can be specified."""
    log.log_assert(len(circuit.nodes.population_names) > 0, "No node population found!")
    if popul_name is None:
        if len(circuit.nodes.population_names) == 1:
            popul_name = circuit.nodes.population_names[0]  # Select the only existing population
        else:
            popul_name = "All"  # Use default name
            log.warning(
                f'Multiple node populations found - Trying to load "{popul_name}" population!'
            )
    log.log_assert(
        popul_name in circuit.nodes.population_names,
        f'Population "{popul_name}" not found in nodes file!',
    )
    nodes = circuit.nodes[popul_name]

    return nodes


def get_edges_population(circuit, popul_name=None, return_popul_name=False):
    """Select default edge population. Optionally, the population name can be specified."""
    #     log.log_assert(len(circuit.edges.population_names) == 1, 'Only a single edge population per file supported!')
    #     edges = circuit.edges[circuit.edges.population_names[0]]
    log.log_assert(len(circuit.edges.population_names) > 0, "No edge population found!")
    if popul_name is None:
        if len(circuit.edges.population_names) == 1:
            popul_name = circuit.edges.population_names[0]  # Select the only existing population
        else:  # Try to use one of default names
            popul_name = "default__default__chemical"
            if popul_name not in circuit.edges.population_names:
                popul_name = "default"
            log.warning(
                f'Multiple edges populations found - Trying to load "{popul_name}" population!'
            )
    log.log_assert(
        popul_name in circuit.edges.population_names,
        f'Population "{popul_name}" not found in edges file!',
    )
    edges = circuit.edges[popul_name]

    if return_popul_name:
        return edges, popul_name
    else:
        return edges


def get_node_positions(nodes, node_ids, vox_map=None):
    """Return x/y/z positions of list of nodes, optionally mapped using VoxelData map."""
    _pop = nodes.to_libsonata
    _sel = libsonata.Selection(node_ids)
    raw_pos = np.column_stack(
        (
            _pop.get_attribute("x", _sel),
            _pop.get_attribute("y", _sel),
            _pop.get_attribute("z", _sel),
        )
    )
    if vox_map:  # Apply voxel map
        pos = vox_map.lookup(raw_pos)
        log.log_assert(
            not np.any(pos == vox_map.OUT_OF_BOUNDS),
            "Out of bounds error in mapped positions from voxel data!",
        )
    else:  # No voxel mapping
        pos = raw_pos
    return raw_pos, pos


def get_connections(edges, pre_ids, post_ids, with_nsyn=False):
    """Returns connections between given src/tgt node IDs, optionally incl. #synapses per connection."""
    it_conns = edges.iter_connections(pre_ids, post_ids, return_edge_count=with_nsyn)
    conns = np.array(
        [([_c.id for _c in _conn[:2]] + list(_conn[2:])) for _conn in it_conns]
    )  # Resolve src/tgt IDs
    if len(conns) == 0:  # Ensure consistent shape even when empty
        return np.empty((0, 2 + int(with_nsyn)))
    return conns


def get_cv_data(data_list, cv_dict=None):
    """Returns training/testing data items of specified cross-validation fold."""
    if cv_dict is None:  # No CV
        return data_list
    else:
        log.log_assert(
            isinstance(cv_dict, dict)
            and "n_folds" in cv_dict
            and "fold_idx" in cv_dict
            and "training_set" in cv_dict,
            'ERROR: Cross-validation "cv_dict" must be a dict containing "n_folds", "fold_idx", and "training_set" keys!',
        )
        log.log_assert(
            cv_dict["n_folds"] > 1 and 0 <= cv_dict["fold_idx"] < cv_dict["n_folds"],
            "ERROR: Cross-validation index error!",
        )

        kf = KFold(n_splits=cv_dict["n_folds"], shuffle=True)
        cv_data_list = []
        for _data in data_list:
            sel_idx = [
                _train_index if cv_dict["training_set"] else _test_index
                for (_train_index, _test_index) in kf.split(_data)
            ][cv_dict["fold_idx"]]
            cv_data_list.append(_data[sel_idx])

        # CV TESTING: Writing CV data splits into data log #
        # dlog_name = f"CVData{cv_dict.get('n_folds', 0)}-{cv_dict.get('fold_idx', -1) + 1}-Train{cv_dict.get('training_set')}"
        # cv_data_dict = {f"cv_data{_i}": _data for _i, _data in enumerate(cv_data_list)}
        # log.data(dlog_name, **cv_data_dict)
        ##############

        return cv_data_list



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os
import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
from bluepysnap import Circuit
from utils import setup_tempdir, TEST_DATA_DIR
from scipy.interpolate import griddata
from voxcell.nexus.voxelbrain import Atlas

import connectome_manipulator.model_building.pos_mapping as test_module


def test_extract():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nrn_pos = circuit.nodes["nodeA"].positions()
    nrn_ids = circuit.nodes["nodeA"].ids()
    flatmap_path = TEST_DATA_DIR

    # Check scaling
    xy_file = "xy_map_ones"
    z_file = "depth_map_ones"
    np.random.seed(0)
    for scale in np.random.rand(10, 3):
        res = test_module.extract(
            circuit, flatmap_path, xy_file, z_file, xy_scale=scale[:2], z_scale=scale[2]
        )
        assert_array_equal(nrn_ids, res["nrn_ids"])
        assert_array_equal(nrn_pos, res["nrn_pos"])
        assert np.all([np.allclose(res["flat_pos"][:, i], scale[i]) for i in range(len(scale))])

    # Check default scaling
    xy_file = "xy_map_ones"
    z_file = "depth_map_ones"
    xymap = Atlas.open(TEST_DATA_DIR).load_data(xy_file)

    res = test_module.extract(circuit, flatmap_path, xy_file, z_file, xy_scale=None, z_scale=None)
    np.testing.assert_allclose(res["flat_pos"][:, 0], xymap.voxel_dimensions[0])
    np.testing.assert_allclose(res["flat_pos"][:, 1], xymap.voxel_dimensions[1])
    np.testing.assert_allclose(res["flat_pos"][:, 2], 1.0)

    # Check interpolation (along z-axis only)
    xy_file = "xy_map_ones"
    z_file = "depth_map_lin"
    dmap = Atlas.open(TEST_DATA_DIR).load_data(z_file)

    ## (a) Nearest-neighbor interpolation
    zidx = dmap.positions_to_indices(nrn_pos.to_numpy(), keep_fraction=False)
    zval = zidx[:, 2].astype(float)  # Z value is same as index in this flatmap
    res = test_module.extract(
        circuit, flatmap_path, xy_file, z_file, xy_scale=[1.0, 1.0], z_scale=1.0, NN_only=True
    )
    assert_array_equal(nrn_ids, res["nrn_ids"])
    assert_array_equal(nrn_pos, res["nrn_pos"])
    assert np.all([np.all(res["flat_pos"][:, i] == 1.0) for i in range(2)])
    assert_array_equal(res["flat_pos"][:, 2], zval)

    ## (b) Linear interpolation using griddata
    zidx = dmap.positions_to_indices(nrn_pos.to_numpy(), keep_fraction=True)
    zval_lin = griddata(np.floor(zidx) + 0.5, np.floor(zidx), zidx)[
        :, 2
    ]  # Data points are at bin centers
    assert np.any(np.isfinite(zval_lin)), "ERROR: Linear interpolation on test data not possible!"
    zval_lin2 = np.interp(
        zidx[:, 2], np.sort(np.floor(zidx[:, 2])) + 0.5, np.sort(np.floor(zidx[:, 2]))
    )  # Interpolate linearly using numpy
    assert_allclose(
        zval_lin[np.isfinite(zval_lin)], zval_lin2[np.isfinite(zval_lin)]
    )  # Check that linearly interpolated along z-axis indeed
    zval[np.isfinite(zval_lin)] = zval_lin[
        np.isfinite(zval_lin)
    ]  # Use lin. interpolation if possible, otherwise keep NN

    res = test_module.extract(
        circuit, flatmap_path, xy_file, z_file, xy_scale=[1.0, 1.0], z_scale=1.0, NN_only=False
    )
    assert_array_equal(nrn_ids, res["nrn_ids"])
    assert_array_equal(nrn_pos, res["nrn_pos"])
    assert np.all([np.all(res["flat_pos"][:, i] == 1.0) for i in range(2)])
    assert_allclose(res["flat_pos"][:, 2], zval)


def test_build():
    np.random.seed(0)
    nrn_ids = np.arange(10)
    flat_pos = 1e3 * np.random.rand(len(nrn_ids), 3)

    model = test_module.build(nrn_ids, flat_pos)
    assert_array_equal(model.apply(gids=nrn_ids), flat_pos)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""connectome_manipulator."""

from connectome_manipulator.version import __version__



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os
import pytest

from bluepysnap import Circuit

from utils import TEST_DATA_DIR
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation
from connectome_manipulator.connectome_manipulation.converters import EdgeWriter
from connectome_manipulator import log


@pytest.fixture
def manipulation():
    m = Manipulation.get("null_manipulation")
    return m


def test_apply(manipulation):
    log.setup_logging()  # To have data logging in a defined state

    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = circuit.edges[circuit.edges.population_names[0]]
    nodes = [edges.source, edges.target]
    tgt_ids = nodes[1].ids()
    edges_table = edges.afferent_edges(tgt_ids, properties=edges.property_names)

    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(tgt_ids)
    res = writer.to_pandas()
    assert res.equals(edges_table)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

from dataclasses import dataclass
from pathlib import Path
import json

from click.testing import CliRunner
from connectome_manipulator.cli import app

from bluepysnap.circuit import Circuit
from libsonata import EdgeStorage, NodeStorage

import pytest

from connectome_manipulator import utils

TEST_DIR = Path(__file__).parent
DATA_DIR = TEST_DIR / "data"
WIRING_CONFIG_TPL = DATA_DIR / "wiring_config.json.tpl"


@dataclass(frozen=True)
class Helper:
    config_file: Path
    output_dir: Path

    @property
    def config(self):
        """Return manipulation config."""
        return utils.load_json(self.config_file)

    @property
    def input_circuit_config(self):
        """Return the input circuit config."""
        return Circuit(Path(self.config["circuit_path"], self.config["circuit_config"])).config

    def run(self, arglist):
        result = CliRunner().invoke(
            cli=app,
            args=arglist,
            catch_exceptions=False,
        )
        print(result.output)
        assert result.exit_code == 0, f"Application exited with non-zero code.\n{result.output}"

    @property
    def output_circuit_config(self):
        """Return the output circuit config."""
        return Circuit(self.output_dir / "circuit_config.json").config

    @property
    def output_edges_file(self):
        """Return path to the output edges file."""
        return self.output_circuit_config["networks"]["edges"][0]["edges_file"]


def _manipulation_config_path(output_dir, config_name):
    """Create and return path to manipulation config."""
    wiring_config = {}
    with open(WIRING_CONFIG_TPL, "r") as config_file:
        wiring_config = json.load(config_file)
    wiring_config["circuit_path"] = str(DATA_DIR)
    wiring_config["circuit_config"] = str(config_name)
    wiring_config["output_path"] = str(output_dir)
    wiring_config_path = output_dir / "wiring_config.json"
    with open(wiring_config_path, "w") as outf:
        json.dump(wiring_config, outf, indent=2)
    return wiring_config_path


@pytest.fixture(scope="module")
def local_connectome__empty_edges(tmpdir_factory):
    """Run the cli command for building the local connectome."""

    output_dir = Path(tmpdir_factory.mktemp("local-connectome-empty-edges"))

    config_file = _manipulation_config_path(output_dir, "circuit_100.json")

    obj = Helper(config_file, output_dir)

    obj.run(
        [
            "manipulate-connectome",
            str(obj.config_file),
            "--output-dir",
            str(obj.output_dir),
            "--overwrite-edges",
            "--convert-to-sonata",
        ],
    )
    return obj


def _check_number_of_edges(edges_file, expected_number):
    edge_storage = EdgeStorage(edges_file)
    pop_name = list(edge_storage.population_names)[0]
    edges = edge_storage.open_population(pop_name)
    assert len(edges) == expected_number, "Wrong number of connections found in output"


def test_number_of_edges__empty_edges(local_connectome__empty_edges):
    """Test the number of edges of the generated edge population."""
    _check_number_of_edges(
        local_connectome__empty_edges.output_edges_file,
        expected_number=134,
    )


def _check_input_output_configs(input_circuit_config, output_circuit_config):
    out_nodes = output_circuit_config["networks"]["nodes"][0]
    out_nodes_file = out_nodes["nodes_file"]

    inp_nodes = input_circuit_config["networks"]["nodes"][0]
    inp_nodes_file = inp_nodes["nodes_file"]

    assert out_nodes_file == inp_nodes_file
    assert Path(out_nodes_file).exists()

    out_edges = output_circuit_config["networks"]["edges"][0]
    out_edges_file = out_edges["edges_file"]

    assert Path(out_edges_file).exists()


def test_generated_circuit_config__empty_edges(local_connectome__empty_edges):
    """Test the validity of the output circuit config."""
    _check_input_output_configs(
        local_connectome__empty_edges.input_circuit_config,
        local_connectome__empty_edges.output_circuit_config,
    )


@pytest.fixture(scope="module")
def local_connectome__existing_edges(tmpdir_factory):
    """Run the cli command for building the local connectome."""

    output_dir = Path(tmpdir_factory.mktemp("local-connectome-existing-edges"))

    config_file = _manipulation_config_path(output_dir, "circuit_100_167.json")

    obj = Helper(config_file, output_dir)
    obj.run(
        [
            "manipulate-connectome",
            str(obj.config_file),
            "--output-dir",
            str(obj.output_dir),
            "--overwrite-edges",
            "--convert-to-sonata",
        ],
    )
    return obj


def test_generated_circuit_config__existing_edges(local_connectome__existing_edges):
    """Test the validity of the output circuit config."""
    _check_input_output_configs(
        local_connectome__existing_edges.input_circuit_config,
        local_connectome__existing_edges.output_circuit_config,
    )


def test_number_of_edges__existing_edges(local_connectome__existing_edges):
    """Test the number of edges of the generated edge population."""
    _check_number_of_edges(
        local_connectome__existing_edges.output_edges_file,
        expected_number=301,
    )
